"""
Utility methods for generating passphrases from a dictionary file of words.
"""

__version_info = (0, 6, 0)
__version__ = '.'.join(map(str, __version_info))

import random
import re
import unicodedata

try:
    unicode
except NameError:
    u = lambda s: s
else:
    u = lambda s: unicode(s)


try:
    from itertools import imap
except ImportError:
    imap = map


# defaults
MIN = 3        # min word length
MAX = 7        # max word length
WORDS = 4      # num words
PAD = ''       # prefix/suffix of passphrase
DELIM = u(' ')  # delimiter
WORD_FILE = '/usr/share/dict/words'


def is_unicode_letter(char):
    """ Answer whether given unicode character is a letter."""
    return unicodedata.category(char) in ('Ll', 'Lu')


def mk_word_matcher(min=MIN, max=MAX, ascii=True):
    """
    Make a word matcher function that accepts a word and returns
    True or False depending on whether the word satisfies the
    the constraints represented by the params.
    :params:
     - min: minimum length of a word
     - max: maximum length of a word
     - ascii: whether to match words that contain only ascii letters
              or words that contain only unicode letters (according to
              ``unicodedata.category``).
    """
    if max < min:
        msg = "min '%s' should be less than or equal to max '%s'"
        raise ValueError(msg % (min, max))
    if ascii:
        pat = re.compile('^[a-zA-Z]{%s,%s}$' % (min, max))

        def matcher(word):
            return bool(pat.match(word))
    else:

        def matcher(word):
            length = len(word)
            return bool(length >= min and length <= max and
                        all(imap(is_unicode_letter, word)))
    return matcher


def get_words(path, min=MIN, max=MAX, ascii=True):
    """
    Get sorted unique words with case normalized to lowercase from file at
    given ``path``, filtering out words that are shorter than ``min`` or longer
    than ``max``. If ``ascii`` is true (default), then only words containing
    just ascii letters are returned. If ``ascii`` is false (and the python
    installation has an appropriate default encoding for the given word file),
    then words in the file that contain only unicode letters (according to
    ``unicodedata.category``) will be included.
    """
    matcher = mk_word_matcher(min=min, max=max, ascii=ascii)
    with open(path) as f:
        words = (line.strip().lower() for line in f)
        if not ascii:
            words = (u(w) for w in words)
        words = list(filter(matcher, set(words)))
    words.sort()
    return words


def sample_words(all_words, k, delim=DELIM, random_case=True):
    """
    Sample ``k`` words from the ``all_words`` word sequence, using ``delim``
    to join the words.

    If ``random_case`` is true (the default), then each word will
    with probability 0.5 be converted to title case.
    """
    all_words = list(all_words)
    words = random.sample(all_words, k)
    if random_case:
        for i, word in enumerate(words):
            if random.choice((True, False)):
                words[i] = word.title()
    return delim.join(words)


def mkpassphrase(path=WORD_FILE, min=MIN, max=MAX, num_words=WORDS,
                 lowercase=False, ascii=True, delim=DELIM, pad=PAD):
    """
    Make a passphrase using given params.

    :params:
    - path: path to a word file, one word per line, encoded with a character
            encoding that is compatible with the python default encoding if
            ``ascii`` is true.
    - min: minimum length of each word in passphrase (at least 1, not greater
           than ``max``).
    - max: maximum length of each word in passphrase (at least 1, not less than
           ``min``).
    - num_words: number of words to include in passphrase (at least 1).
    - lowercase: whether to keep initial letter of each word lowercased or
                 capitalize it with probability 0.5.
    - ascii: whether to only include words contain just ascii letters, or to
             also include words that contain any unicode letter (but no
             characters that are not unicode letters).
    - delim: the delimiter to use for joining the words in the passphrase.
    - pad: a string to use as a prefix and suffix of the generated passphrase.

    """
    if min < 1:
        raise ValueError("'min' must be at least 1")
    if max < 1 or max < min:
        raise ValueError("'max' must be positive and greater than"
                         " or equal to 'min'")
    if num_words < 1:
        raise ValueError("'num_words' must be at least 1")

    all_words = get_words(path, min=min, max=max, ascii=ascii)
    passphrase = sample_words(all_words, num_words, delim=delim,
                              random_case=not lowercase)
    passphrase = pad + passphrase + pad
    num_candidates = len(all_words)
    if not lowercase:
        num_candidates *= 2
    return passphrase, num_candidates


def num_possible(num_candidates, num_words):
    """
    Calculate the number of possible passphrases that could be generated
    by selecting ``num_words`` from ``num_candidates``, each of which
    must be at least 1.
    """
    if num_candidates < 1:
        raise ValueError('num_candidates must be positive')
    if num_words < 1:
        raise ValueError('num_words must be positive')

    n, k = num_candidates, num_words
    possible = 1
    while k > 0:
        possible *= n
        n -= 1
        k -= 1
    return possible



import sys
import os
import argparse

import mkpassphrase as M


def main(argv=None):
    if argv is None:
        argv = sys.argv
    parser = argparse.ArgumentParser(description='Generate a passphrase.')
    parser.add_argument('-n', '--num-words', type=int, metavar='NUM_WORDS',
                        help='Number of words in passphrase', default=M.WORDS)
    parser.add_argument('--min', type=int, metavar='MIN',
                        help='Minimum word length', default=M.MIN)
    parser.add_argument('--max', type=int, metavar='MAX',
                        help='Maximum word length', default=M.MAX)
    parser.add_argument('-f', '--word-file', type=str, metavar='WORD_FILE',
                        help='Word file path (one word per line)',
                        default=M.WORD_FILE)
    parser.add_argument('--lowercase', action='store_true',
                        dest='lowercase', default=False,
                        help='Make each word entirely lowercase, rather than'
                             ' the default behavior of choosing Titlecase or'
                             ' lowercase for each word (with probability 0.5)')
    parser.add_argument('--non-ascii', action='store_false',
                        dest='ascii', default=True,
                        help='Whether to allow words with non-ascii letters')
    parser.add_argument('-p', '--pad', metavar='PAD', default='',
                        help='Pad passphrase using PAD as prefix and suffix')
    parser.add_argument('-d', '--delimiter', dest='delimiter', default=' ',
                        metavar='DELIM',
                        help='Use DELIM to separate words in passphrase')
    parser.add_argument('-V', '--version', action='store_true',
                        help="Show version")

    args = parser.parse_args()
    if args.version:
        print("%s %s" % (M.__name__, M.__version__))
        sys.exit(0)
    if args.min > args.max:
        parser.exit("--max must be equal to or greater than --min")
    if args.min < 1 or args.max < 1:
        parser.exit('--min and --max must be positive')
    if args.num_words < 1:
        parser.exit('--num-words must be positive')
    if not os.access(args.word_file, os.R_OK):
        parser.exit("word file does not exist or is not readable: %s" %
                    args.word_file)

    passphrase, num_candidates = M.mkpassphrase(
        path=args.word_file,
        min=args.min,
        max=args.max,
        num_words=args.num_words,
        lowercase=args.lowercase,
        delim=args.delimiter,
        pad=args.pad
    )
    possibilities = M.num_possible(num_candidates, args.num_words)

    print(passphrase)
    print("{0:,g} unique candidate words".format(num_candidates))
    print("{0:,g} possible passphrases".format(possibilities))


if __name__ == '__main__':
    main()
